static char rcsid[] = "@(#)$Id: mime_types.c,v 1.10 2006/09/15 19:56:31 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.10 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@posti.FMI.FI> 
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/


#include "def_misc.h"
#include "s_me.h"

DEBUG_VAR(Debug,__FILE__,"mime");

#include <errno.h>
#ifndef ANSI_C
extern int errno;
#endif

static unsigned char * s2us P_((char *str));
static unsigned char * s2us(str)
     char *str;
{
    return (unsigned char *)str;
}

struct mime_types_item {
    char                * extension;
    struct media_type   * type;

    char                * params;
    charset_t           params_cs;

    /* magic */
    int                 offset;
    int                 length;
    unsigned char     * bytes;

};

/* -------------------------------------------------------------------- */

#define U_(x) (unsigned char *)(x)

static struct mime_types_item builtin_mimetypes_map_0[] = {


    /* application/postscript */
    { "ps", &application_postscript, NULL, NULL,   0,4, U_("%!PS") },

    /* image/gif */
    { "gif", &image_gif,             NULL, NULL,   0,3, U_("GIF") },

    /* image/jpeg                             magic includes \0  */
    { "jpeg",  &image_jpeg,          NULL, NULL,   6,5, U_("JFIF") },
    { "jpg",  &image_jpeg,           NULL, NULL,   6,5, U_("JFIF") },

    /* image/tiff */
    { "tiff",  &image_tiff,          NULL, NULL,   0,4, U_("MM\x00\x2a") },
    { "tiff",  &image_tiff,          NULL, NULL,   0,4, U_("II\x2a\x00") },
    { "tif",   &image_tiff,          NULL, NULL,   0,4, U_("MM\x00\x2a") },
    { "tif",   &image_tiff,          NULL, NULL,   0,4, U_("II\x2a\x00") },


    { NULL, NULL,                    NULL, NULL,   0,0,NULL },
};

struct mime_types_item  * builtin_mimetypes_map = builtin_mimetypes_map_0;


static int ok_byte P_((int byte));
static int ok_byte(byte)
     int byte;
{
    if (!isascii(byte))
	return 0;
    if (iscntrl(byte))
	return 0;
    if (!isprint(byte))
	return 0;
    if ('"' == byte)
	return 0;
    return 1;
}

static void dump_magic P_((FILE *f, int offset, int length,
			   unsigned char * bytes));
static void dump_magic(f,offset,length,bytes)
     FILE * f;
     int                 offset;
     int                 length;
     unsigned char     * bytes;
{
    int q = 0;
    int i;

    fprintf(f,">%d,",offset);
    
    for (i = 0; i < length; i++) {

	if (!q && i < length-1 &&
	    ok_byte(bytes[i]) &&
	    ok_byte(bytes[i+1])) {

	    fputc(' ',f);
	    fputc('"',f);
	    q = 1;
	}

	if (!q) 
	    fputc(' ',f);
	else {
	    if (! ok_byte(bytes[i])) {
		fputc('"',f);
		q = 0;	    
		fputc(' ',f);			    
	    }
	}

	if (q)
	    fputc(bytes[i],f);
	else
	    fprintf(f,"%d",bytes[i]);
    }

    if (q) 
	fputc('"',f);

    fputc('\n',f);
}


static int parse_magic P_((char * buf, int l1, int *reslen,  
			   char **resbuffer));
static int parse_magic(buf,l1,reslen,resbuffer)
     char * buf; 
     int l1; 
     int *reslen;  
     char **resbuffer;
{
    char * buffer = NULL;
    int    len = 0;
    char * p = buf;

    while (p < buf +l1) {
   
	if (' ' == *p) {
	    p++;
	} else if ('"' == *p) {
	    char * start = ++p;
	    int len1;

	    while (p < buf +l1) {
		if ('"' == *p) 
		    break;
		p++;
	    }
	    if ('"' != *p) {
		DPRINT(Debug,2,(&Debug,"parse_magic: fail on: %s\n",p));
		goto fail;
	    }
	    len1 = p-start;
	    p++;

	    if (0 == len1) {
		DPRINT(Debug,2,(&Debug,"parse_magic: fail on: %s\n",p));
		goto fail;
	    }
	    
	    buffer= safe_realloc(buffer,len+len1);
	    memcpy(buffer+len,start,len1);
	    len += len1;
	} else {
	    char * x;

	    int L = strtol(p,&x,0);

	    if ('\0' != *x && 
		' ' != *x) {
		DPRINT(Debug,2,(&Debug,"parse_magic: fail on: %s\n",p));
		goto fail;
	    }

	    if (L < 0) {
		DPRINT(Debug,2,(&Debug,"parse_magic: fail on: %s\n",p));
		goto fail;
	    }

	    p = x;


	    if (L < 256) {
		buffer= safe_realloc(buffer,len+1);
		buffer[len] = L;
		len++;
	    } else if (L <= 0x10000 &&
		       sizeof (uint16) == 2) {
		/* Not well defined -- host byte order */
		uint16 s = L;

		buffer= safe_realloc(buffer,len+2);
		memcpy(buffer+len,&s,2);
		len += 2;
		
	    } else {   /* Undefined :-) */
		buffer= safe_realloc(buffer,len+sizeof L);
		memcpy(buffer+len,&L,sizeof L);
		len += sizeof L;
	    }
	}
    }

    if (len > 0) {       
	*reslen    = len;
	*resbuffer = buffer;
	return 1;
    }
    DPRINT(Debug,2,(&Debug,"parse_magic: fail -- nothing collected: %s\n",
		    buf));

 fail:
    if (buffer)
	free(buffer);

    *reslen    = 0;
    *resbuffer = NULL;
    return 0;   
}

struct mime_types_item  * load_mime_types_map(filename,errors)
     const char *filename;
     int *errors;
{
    struct mime_types_item * result;
    charset_t  cs  = system_charset;    /* Charset of file */

    int result_len = 0;
    FILE * f;
    int max_result = 0;
    int c;
    char * buf = NULL;
    int err = can_open(filename,"r");

    if (err) {
	DPRINT(Debug,2,(&Debug,"load_mime_types_map: %s: %s (can_open)\n",
			filename,error_description(err)));
	return NULL;
    }

    f = fopen(filename,"r");
    if (!f) {
 	int err = errno;
	DPRINT(Debug,2,(&Debug,"load_mime_types_map: %s: %s\n",
			filename,error_description(err)));
	return NULL;
    }

    while(EOF != (c = fgetc(f)))
	if ('\n' == c)
	    max_result++;

    DPRINT(Debug,10,(&Debug,"load_mime_types_map: %s, max_result=%d\n",
		     filename,max_result));
    
    if (!max_result) {
	fclose(f);
	return NULL;
    }
    rewind(f);
    
    result = safe_malloc((max_result +1) * sizeof (result[0]));
    
    /* malloc_gets reallocates buf and do not return \n.
       It returns -1 if line is longer than given limit (32000)
       and -2 is returned on error  (buffer may still be alloced)
    */

    while (result_len < max_result &&
	   !feof(f) && !ferror(f)) {
	int l1 = malloc_gets(&buf,32000,f);
	char * c = buf;
	char * extension = NULL;
	char * mime_type = NULL;
	char * subtype   = NULL;
	char * params    = NULL;
	char * c1;
	media_type_t TYPE;

	int ch;

	if (-1 == l1) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeTooLongLine,
			      "%.30s: Too long line: %.30s..."),
		      filename,buf ? buf : "???");
	    (*errors) ++;
	    goto OUT;

	} else if (-2 == l1) {
	    int err = errno;
	    
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeReadError,
			  "%.30s: Reading error: %s"),
		      filename,error_description(err));
	    (*errors) ++;	    

	    goto OUT;
	} else if (0 == l1)
	    continue;

	while (l1-- > 0 && whitespace(buf[l1]))
	    buf[l1] = '\0';

	
	while (*c && whitespace ( *c)) /* skip leading whitespace */
	    c++;
	if (*c == '#') /* Skip comments */
	    continue;
	if (! *c)
	    continue;

	/* magic on wrong place */
	if ('>' == *c) {
            lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
                              "%.30s: Bad line: %.30s..."),
                      filename,buf);
            (*errors) ++;
            break;
	}

	c1 = strpbrk(c," \t");
        if (!c1) {
            lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
                              "%.30s: Bad line: %.30s..."),
                      filename,buf);
            (*errors) ++;
            break;
        }
        *c1 = '\0';

	if (0 == strcmp(c,"-"))
	    extension = NULL;
	else
	    extension = c;

	c1++;

	while (*c1 && whitespace (*c1)) /* skip leading whitespace */
            c1++;
        if (!*c1) {
            lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
                              "%.30s: Bad line: %.30s..."),
                      filename,buf);
            (*errors) ++;
            break;
        }
	
	mime_type = c1;

	while (*c1) {
	    if ('/' == *c1) {
		subtype = c1+1;
		*c1 = '\0';
	    }
	    if (';' == *c1)
		break;
	    if (whitespace (*c1)) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
				  "%.30s: Bad line: %.30s..."),
			  filename,buf);
		(*errors) ++;
		goto OUT;
	    }
	    c1++;
	}

	if (';' == *c1) {
	    *c1 = '\0';

	    c1++;
	    
	    while (*c1 && whitespace (*c1)) /* skip leading whitespace */
		c1++;

	    if (*c1)
		params = c1;

	} else if ('\0' != *c1) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
			      "%.30s: Bad line: %.30s..."),
		      filename,buf);
	    (*errors) ++;
	    goto OUT;	    
	}

	if (! mime_type[0] || !subtype || !subtype[0]) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
			      "%.30s: Bad line: %.30s..."),
		      filename,buf);
	    (*errors) ++;
	    goto OUT;
	}

	TYPE = give_media_type(mime_type,subtype,1);
	if (!TYPE) {
	    DPRINT(Debug,1,(&Debug,
			    "load_mime_types_map: %s: FAILED to get type %s/%s\n",
			    filename,mime_type,subtype));
	    continue;
	}

	result[result_len].extension   = extension ?
	    safe_strdup(extension) : NULL;
	result[result_len].type        = TYPE;
	result[result_len].params      = params ?
	    safe_strdup(params) : NULL;
	result[result_len].params_cs    = cs;

	result[result_len].offset = 0;
	result[result_len].length = 0;
	result[result_len].bytes  = NULL;

	/* magic */
	ch = fgetc(f);

	if (ch == '>') {
	    int offset = 0;
	    char * p;
	    char * buffer = NULL;
	    int    len    = 0;

	    l1 = malloc_gets(&buf,32000,f);


	    if (-1 == l1) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeTooLongLine,
				  "%s: Too long line: %.30s..."),
			  filename,buf ? buf : "???");
		(*errors) ++;

		goto CLEAR;
		
	    } else if (-2 == l1) {
		int err = errno;
		
		lib_error(CATGETS(elm_msg_cat, MeSet, MeReadError,
				  "%s: Reading error: %s"),
			  filename,error_description(err));
		(*errors) ++;	    

		goto CLEAR;

	    } else if (0 == l1)
		goto nomagic;


	    p = buf;
	    if ('"' == *buf) {
		
		if (! parse_magic(buf,l1,&len,&buffer)) {
		    lib_error(CATGETS(elm_msg_cat, MeSet, MeBadMagic,
				      "%s: bad magic: %.30s..."),
			      filename,buf ? buf : "???");

		    (*errors) ++;	    
		    goto CLEAR;
		}
	    
	    } else {
		long L = strtol(buf,&p,10);
	    
		if (L < 0 || L > 1024 || *p != ',') {
		    lib_error(CATGETS(elm_msg_cat, MeSet, MeMagicOffset,
				      "%s: bad magic offset: %.30s..."),
			      filename,buf ? buf : "???");

		    (*errors) ++;	    		
		    goto CLEAR;
		}
		offset = L;
		p++;

		if (! parse_magic(p,l1 - (p-buf),&len,&buffer)) {
		    lib_error(CATGETS(elm_msg_cat, MeSet, MeBadMagic,
				      "%s: bad magic: %.30s..."),
			      filename,buf ? buf : "???");

		    (*errors) ++;	    
		    goto CLEAR;
		}

	    }		
	    
	    result[result_len].offset = offset;
	    result[result_len].length = len;
	    result[result_len].bytes  = s2us(buffer);
		
	} else if (EOF != ch) {
	    ungetc(ch,f);

	nomagic:
	    if (!extension) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeMagicExtension,
				  "%.30s: magic or extension required: %.30s..."),
			  filename,buf ? buf : "???");
		(*errors) ++;	    		


	    CLEAR:
		if (result[result_len].extension) {
		    free(result[result_len].extension);
		    result[result_len].extension = NULL;
		}
		if (result[result_len].params) {
		    free(result[result_len].params);
		    result[result_len].params = NULL;
		}
		goto OUT;
	    }		
	}

	result_len++;
    }

    if (ferror(f)) {
	int err = errno;

	lib_error(CATGETS(elm_msg_cat, MeSet, MeReadError,
			  "%s: Reading error: %s"),
		  filename,error_description(err));
	(*errors) ++;
    }
    

 OUT:
    if (buf) {
	free(buf);
    }

    result[result_len].extension   = NULL;
    result[result_len].type         = NULL;
    result[result_len].params       = NULL;
    result[result_len].params_cs    = NULL;

    if (!feof(f)) {
	DPRINT(Debug,3,(&Debug,
			"load_mime_types_map: %s, All results not readed\n",
			filename));
    }

    fclose(f);

    
    DPRINT(Debug,10,(&Debug,
		     "load_mime_types_map: %s, result_len=%d\n",
		     filename,result_len));
    
    return result;
}

struct mime_types_item * loc_mime_type(list,extension)
     struct mime_types_item *list;
     char *extension;
{
    struct mime_types_item *ptr;

    for (ptr = list; ptr->type; ptr++) {
	if (ptr->extension &&
	    0 == istrcmp(ptr->extension,extension) &&
	    0 == ptr->length)
	    return ptr;
    }

    return NULL;
}

media_type_t mime_type_to_media_type(item)
     CONST struct mime_types_item *item;
{
    return item->type; 
}

CONST char * mime_type_to_params(item,cs)
     CONST struct mime_types_item *item;
     charset_t *cs;
{
    if (cs)
	*cs = item->params_cs;

    return item->params;
}

CONST char * mime_type_to_extension(item)
     struct mime_types_item *item;
{
    return item->extension;
}



void dump_mime_types_map(f,map)
     FILE *f; 
     struct mime_types_item  *map;
{
    struct mime_types_item  *ptr;

    for (ptr = map; ptr && ptr->type; ptr++) {

	if (ptr->extension)	    
	    fputs(ptr->extension,f);
	else
	    fputc('-',f);
	fputc('\t',f);
	fputs(get_major_type_name(ptr->type),f);
    	fputc('/',f);
	fputs(get_subtype_name(ptr->type),f);
	
	if (ptr->params) {
	    fputc(';',f);
	    fputc(' ',f);
	    fputs(ptr->params,f);
	}
    	fputc('\n',f);

	if (ptr->bytes) {
	    dump_magic(f,ptr->offset,ptr->length,ptr->bytes);
	}
    }
}

static int match_item P_((struct mime_types_item *p1,
			  struct mime_types_item *p2));
static int match_item(p1,p2)
     struct mime_types_item *p1;
     struct mime_types_item *p2;
{
    if (p1->extension && p2->extension) {
	if (0 != strcmp(p1->extension,p2->extension))
	    return 0;
    } else if (p1->extension || p2->extension)
	return 0;

    if (p1->length != p2->length)
	return 0;
    if (p1->offset != p2->offset)
	return 0;

    if (p1->bytes && p2->bytes) {
	if (0 != memcmp(p1->bytes,p2->bytes,p1->length))
	    return 0;
    } else if (p1->bytes || p2->bytes) 
	return 0;

    return 1;
}

void change_mime_types_map(map,new) 
     struct mime_types_item **map;
     struct mime_types_item *new;
{
    int count = 0;

    struct mime_types_item *ptr;

    for (ptr = *map; ptr && ptr->type; ptr++)
	count++;

    for (ptr = new; ptr && ptr->type; ptr++)
	count++;

    if (!count)
	return;

    if (!*map) {
	*map = safe_malloc((count+1) * sizeof ((*map)[0]));
	(*map)->type   = NULL;
	
    } else
	*map = safe_realloc(*map, (count+1) * sizeof ((*map)[0]));


   for (ptr = new; ptr && ptr->type; ptr++) {
	struct mime_types_item *ptr2;

	for (ptr2 = *map; ptr2->type; ptr2++) {
	    if (match_item(ptr,ptr2)) {
		goto set_it;
	    }
	}

	if (ptr2 > (*map) + count)
	    panic("CHARSET PANIC",__FILE__,__LINE__,"change_mime_types_map",
		  "Overflow",0);
	
	ptr2->extension = NULL;
	if (ptr->extension)
	    ptr2->extension = safe_strdup(ptr->extension);

	ptr2->length = ptr->length;
	ptr2->offset = ptr->offset;

	ptr2->bytes = NULL;
	if (ptr->bytes) {
	    ptr2->bytes = safe_malloc(ptr->length);
	    memcpy(ptr2->bytes,ptr->bytes,ptr->length);
	}
	ptr2->params = NULL;

	(ptr2+1)->type   = NULL;

   set_it:
	ptr2->type = ptr->type;
	ptr2->params_cs = ptr->params_cs;

	if (ptr->params) {
	    ptr2->params = strmcpy(ptr2->params,ptr->params);

	} else if (ptr2->params) {
	    free(ptr2->params);
	    ptr2->params = NULL;
	}
	
   }
}


void free_mime_types_map(map)
     struct mime_types_item **map;
{
    struct mime_types_item *ptr;

    for (ptr = *map; ptr && ptr->type; ptr++) {

	if (ptr->extension) {
	    free(ptr->extension);
	    ptr->extension = NULL;
	}

	if (ptr->params) {
	    free(ptr->params);
	    ptr->params = NULL;
	}

	if (ptr->bytes) {
	    free(ptr->bytes);
	    ptr->bytes = NULL;
	}

    }



    free(*map);
    *map = NULL;
}



/* -------------------------------------------------------------------- */

#define SCAN_LIST_magic		0xFC05

struct scan_list {
    unsigned short         magic;  /* SCAN_LIST_magic */

    int current_ptr;
    
    struct scan_list_item {
	int matches;     /* 0  == scan
			    -1 == fail
			    1  == match */

	struct mime_types_item *item;
    } * list;
    int list_len;    
};

struct scan_list * get_scan_list(list,extension)
     struct mime_types_item *list;
     char *extension;
{
    struct mime_types_item *ptr;

    struct scan_list *ret = safe_malloc(sizeof (*ret));
    
    /* bzero is defined hdrs/defs.h */
    bzero((void *)ret,sizeof (*ret));

    ret->magic         = SCAN_LIST_magic;
    ret->current_ptr   = 0;

    ret->list      = NULL;
    ret->list_len  = 0;

    for (ptr = list; ptr->type; ptr++) {
	if (0 == ptr->length)
	    continue;

	if (! ptr->extension ||
	    extension && ptr->extension &&
	    0 == istrcmp(ptr->extension,extension)) {

	    ret->list = safe_realloc(ret->list,
				     sizeof (ret->list[0]) *
				     (ret->list_len +1));

	    ret->list[ret->list_len].matches = 0;
	    ret->list[ret->list_len].item    = ptr;

	    ret->list_len++;
	}
    }   

    DPRINT(Debug,3,(&Debug,"get_scan_list: %d match (extension %s)\n",
		    ret->list_len,extension ? extension : "<none>"));
    
    return ret;
}

/* If there is type without magic, it means that any content
   is accepted for type */
struct scan_list * get_scan_list_by_type(list,mime_type)
     struct mime_types_item *list;
     media_type_t mime_type;
{
    struct mime_types_item *ptr;

    struct scan_list *ret = safe_malloc(sizeof (*ret));
    
    /* bzero is defined hdrs/defs.h */
    bzero((void *)ret,sizeof (*ret));

    ret->magic         = SCAN_LIST_magic;
    ret->current_ptr   = 0;

    ret->list      = NULL;
    ret->list_len  = 0;

    for (ptr = list; ptr->type; ptr++) {

	if (ptr->type == mime_type) {
	    ret->list = safe_realloc(ret->list,
				     sizeof (ret->list[0]) *
				     (ret->list_len +1));

	    ret->list[ret->list_len].matches = 0;
	    ret->list[ret->list_len].item    = ptr;

	    ret->list_len++;

	}
    }

    DPRINT(Debug,3,(&Debug,"get_scan_list_by_type: %d match (type %s/%s)\n",
		    ret->list_len,
		    get_major_type_name(mime_type),
		    get_subtype_name(mime_type)));
	   
    return ret;
}



void free_scan_list(list)
     struct scan_list **list;
{

    if (SCAN_LIST_magic != (*list)->magic)
	panic("MIME TYPES PANIC",__FILE__,__LINE__,"free_scan_list",
	      "Bad magic",0);

    if ((*list)->list) {
	free((*list)->list);
	(*list)->list = NULL;
    }
    (*list)->list_len = 0;

    (*list)->magic = 0;  /* Invalidate */

    free(*list);
    *list = NULL;
}

void append_scanlist(list,other)
     struct scan_list *list;
     struct scan_list *other;
{
    int newlen,i;

    if (SCAN_LIST_magic != list->magic)
	panic("MIME TYPES PANIC",__FILE__,__LINE__,"append_scanlist",
	      "Bad magic",0);

    if (SCAN_LIST_magic != other->magic)
	panic("MIME TYPES PANIC",__FILE__,__LINE__,"append_scanlist",
	      "Bad magic",0);

    if (list->current_ptr)
	panic("MIME TYPES PANIC",__FILE__,__LINE__,"append_scanlist",
	      "Called on middle on scan",0);

    /* Avoid allocing 0 bytes */
    if (other->list_len < 1)
	return;
	

    newlen = list->list_len + other->list_len;
    
    list->list = safe_realloc(list->list,
			     sizeof (list->list[0]) * newlen);
			     
    for (i = 0; i < other->list_len; i++) {

	if (list->list_len >= newlen)
	    panic("MIME TYPES PANIC",__FILE__,__LINE__,"append_scanlist",
		  "Overflow",0);

	list->list[list->list_len].matches = 0;
	list->list[list->list_len].item    = other->list[i].item;
	
	list->list_len++;
    }
}


void filter_scanlist(list,att,filter)
     struct scan_list *list; 
     const mime_t * att;
     filter_scanlist_OK_f filter;
{
    int i,p;
    int oldlen;

    if (SCAN_LIST_magic != list->magic)
	panic("MIME TYPES PANIC",__FILE__,__LINE__,"filter_scanlist",
	      "Bad magic",0);

    if (list->current_ptr)
	panic("MIME TYPES PANIC",__FILE__,__LINE__,"filter_scanlist",
	      "Called on middle on scan",0);

    for (i = 0, p = 0; i < list->list_len; i++) {

	if (filter(att,list->list[i].item)) {

	    if (p != i) {

		DPRINT(Debug,7,(&Debug,
				"filter_scanlist: Moving %d -> %d\n",
				i,p));

		
		list->list[p] = list->list[i];

	    }

	    p++;
	} else {

	    DPRINT(Debug,7,(&Debug,
			    "filter_scanlist: Filtering item %d\n",
			    i));
	
	}
    }

    oldlen = list->list_len;

    if (p != list->list_len) {

	DPRINT(Debug,7,(&Debug,
			"filter_scanlist: Shortening list %d => %d\n",
			list->list_len,p));

	list->list_len = p;
    }

    for (; p < list->list_len; p++) {

	DPRINT(Debug,7,(&Debug,
			"filter_scanlist: Zeroing old item %d\n",
			p));
    

	list->list[p].item = NULL;
    }

}


int scanlist_need_more(list,ch)
     struct scan_list *list; 
     int ch;
{
    int i;
    int needmore = 0;


    if (SCAN_LIST_magic != list->magic)
	panic("MIME TYPES PANIC",__FILE__,__LINE__,"scanlist_need_more",
	      "Bad magic",0);

    if (0 == list->current_ptr) {
	DPRINT(Debug,7,(&Debug,
			"scanlist_need_more: %d items on scan\n",
			list->list_len));
    }

    
    for (i = 0; i < list->list_len; i++) {
	
	if (0 == list->list[i].matches) {
	    int x;

	    struct mime_types_item *item = 
		list->list[i].item ;

	    if (list->current_ptr < item->offset) {
		needmore = 1;
		continue;
	    }

	    if (item->offset == list->current_ptr && 0 == item->length) {
		DPRINT(Debug,7,(&Debug,
				"scanlist_need_more: [%d] no magic -- matches to any content at offset %d\n",
				i,list->current_ptr));
		
		list->list[i].matches = 1;

		goto skip_scan;

	    } else if (list->current_ptr >= item->offset + item->length) 
		panic("MIME TYPES PANIC",__FILE__,__LINE__,
		      "scanlist_need_more",
		      "matches was not set",0);

	    x = list->current_ptr - item->offset;
	    
	    if (ch == item->bytes[x]) {
		/* Still matches */


		if (x == item->length-1) {

		    int j;

		    list->list[i].matches = 1;

		    DPRINT(Debug,7,(&Debug,
				    "scanlist_need_more: [%d] matches on offset %d ch=%d\n",
				    i,list->current_ptr,ch));
		    
		    DPRINT(Debug,7,(&Debug,
				    "scanlist_need_more: pattern %d,",
				    item->offset));

		    for (j = 0; j < item->length; j++) {
			DPRINT(Debug,7,(&Debug," %d",item->bytes[j]));
		    }
		    DPRINT(Debug,7,(&Debug,"\n"));

		}

	    } else {
		int j;

		list->list[i].matches = -1;

		DPRINT(Debug,7,(&Debug,
				"scanlist_need_more: [%d] fails on offset %d ch=%d\n",
				i,list->current_ptr,ch));
		
		DPRINT(Debug,7,(&Debug,
				"scanlist_need_more: pattern %d,",
				item->offset));
		
		for (j = 0; j < item->length; j++) {
		    DPRINT(Debug,7,(&Debug," %d",item->bytes[j]));
		}
		DPRINT(Debug,7,(&Debug,"\n"));

	    }
	}
	
    skip_scan:
	if (0 == list->list[i].matches)
	    needmore = 1;
    }

    list->current_ptr++;
    
    return needmore;
}

/* Returns 1 is list is not empty */
int have_scanlist(list)
     struct scan_list *list; 
{
    if (SCAN_LIST_magic != list->magic)
	panic("MIME TYPES PANIC",__FILE__,__LINE__,"have_scanlist",
	      "Bad magic",0);

    return list->list_len != 0;
}

struct mime_types_item * loc_mime_type_from_scan(list)
     struct scan_list *list; 
{
    struct mime_types_item *ret = NULL;
    int match_len  = -1;
    int i;

    if (SCAN_LIST_magic != list->magic)
	panic("MIME TYPES PANIC",__FILE__,__LINE__,"scanlist_need_more",
	      "Bad magic",0);

    /* Pick longest match */
    
    for (i = 0; i < list->list_len; i++) {

	if (1 == list->list[i].matches &&
	    match_len < list->list[i].item->length) {

	    ret = list->list[i].item;
	    match_len = list->list[i].item->length;

	}
    }
    
    return ret;
}

struct mime_types_item * loc_mime_type_from_scan_extension(list,extension)
     struct scan_list *list; 
     char *extension;
{
    struct mime_types_item *ret = NULL;
    int match_len  = -1;
    int i;

    if (SCAN_LIST_magic != list->magic)
	panic("MIME TYPES PANIC",__FILE__,__LINE__,"scanlist_need_more",
	      "Bad magic",0);

    /* Pick longest match -- first try matching extension */
    
    for (i = 0; i < list->list_len; i++) {

	if (1 == list->list[i].matches &&
	    match_len < list->list[i].item->length &&
	    list->list[i].item->extension &&
	    0 == istrcmp(list->list[i].item->extension,extension)
	    
	    ) {

	    ret = list->list[i].item;
	    match_len = list->list[i].item->length;

	}
    }
    
    if (!ret)
	ret = loc_mime_type_from_scan(list);

    return ret;
}



/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
